from sqlalchemy.connectors.pyodbc import PyODBCConnector

from databricks_dbapi import odbc
from databricks_dbapi.sqlalchemy_dialects.base import DatabricksDialectBase


class DatabricksPyodbcDialect(DatabricksDialectBase, PyODBCConnector):
    name = b"databricks"
    driver = b"pyodbc"

    @classmethod
    def dbapi(cls):
        return odbc

    def create_connect_args(self, url):
        elements, kwargs = super().create_connect_args(url=url)
        # we use user (following PEP249 guidelines), pyodbc uses username
        kwargs["user"] = kwargs.pop("username")
        return elements, kwargs



"""Provide a function to create an ODBC connection to a Databricks cluster."""
import pyodbc
from pyodbc import *  # Make globals and exceptions visible in this module per PEP 249


def connect(
    host,
    port=443,
    database="default",
    http_path=None,
    token=None,
    user=None,
    password=None,
    driver_path=None,
    **kwargs,
):
    """Create an ODBC DBAPI connection to a Databricks workspace or SQL Analytics cluster.

    Create a DBAPI connection to a Databricks cluster, which can be used to generate
    DBAPI cursor(s). Provide an ``http_path`` from the cluster's
    JDBC/ODBC connection details.

    For authentication, provide either a ``token`` OR both a ``user`` and ``password``.
    Token authentication is strongly recommended over passwords.

    The simplest connection requires providing args ``host``, ``http_path``,
    ``token``, and the ODBC ``driver_path``.  The default path on Mac OSX is likely:
    ``/Library/simba/spark/lib/libsparkodbc_sbu.dylib``

    :param str host: the server hostname from the cluster's JDBC/ODBC connection page.
    :param int port: the port number from the cluster's JDBC/ODBC connection page.
    :param str database: the database to use
    :param str http_path: the HTTP Path as shown in the cluster's JDBC/ODBC connection
        page. Required if using Azure platform.
    :param str token: a Databricks API token.
    :param str user: a Databricks user name.
    :param str password: the corresponding Databricks user's password.
    :param str driver_path: the absolute path to the ODBC driver.
    :param dict kwargs: keyword args passed to ``pyodbc.connect``
    """
    if driver_path is None:
        raise ValueError("Driver path must be provided.")

    if token is not None:
        user = "token"
        password = token
    elif user is not None and password is not None:
        pass
    else:
        raise ValueError("Missing arguments. Must provide either token or user/password.")

    connection_string = (
        f"Driver={driver_path};Database={database};Host={host};"
        f"Port={port};httpPath={http_path};UID={user};PWD={password};"
        "transportMode=http;ssl=1;AuthMech=8;SparkServerType=3;ThriftTransport=1"
    )

    # autocommit is required
    return pyodbc.connect(connection_string, autocommit=True, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['databricks_dbapi', 'databricks_dbapi.sqlalchemy_dialects']

package_data = \
{'': ['*']}

install_requires = \
['pyhive[hive]>=0.6.1,<0.7.0', 'pyodbc>=4.0.30,<5.0.0']

extras_require = \
{'sqlalchemy': ['sqlalchemy>=1.3,<2.0']}

entry_points = \
{'sqlalchemy.dialects': ['databricks.pyhive = '
                         'databricks_dbapi.sqlalchemy_dialects.hive:DatabricksPyhiveDialect',
                         'databricks.pyodbc = '
                         'databricks_dbapi.sqlalchemy_dialects.odbc:DatabricksPyodbcDialect']}

setup_kwargs = {
    'name': 'databricks-dbapi',
    'version': '0.5.0',
    'description': 'A DBAPI 2.0 interface and SQLAlchemy dialect for Databricks interactive clusters.',
    'long_description': 'databricks-dbapi\n================\n\n|pypi| |pyversions|\n\n.. |pypi| image:: https://img.shields.io/pypi/v/databricks-dbapi.svg\n    :target: https://pypi.python.org/pypi/databricks-dbapi\n\n.. |pyversions| image:: https://img.shields.io/pypi/pyversions/databricks-dbapi.svg\n    :target: https://pypi.python.org/pypi/databricks-dbapi\n\nA thin wrapper around `pyhive <https://github.com/dropbox/PyHive>`__ and `pyodbc <https://github.com/mkleehammer/pyodbc>`__ for creating a `DBAPI <https://www.python.org/dev/peps/pep-0249/>`__ connection to Databricks Workspace and SQL Analytics clusters. SQL Analytics clusters require the `Simba ODBC driver <https://databricks.com/spark/odbc-driver-download>`__.\n\nAlso provides SQLAlchemy Dialects using ``pyhive`` and ``pyodbc`` for Databricks clusters. Databricks SQL Analytics clusters only support the ``pyodbc``-driven dialect.\n\nInstallation\n------------\n\nInstall using pip:\n\n.. code-block:: bash\n\n    pip install databricks-dbapi\n\n\nFor SQLAlchemy support install with:\n\n.. code-block:: bash\n\n    pip install databricks-dbapi[sqlalchemy]\n\nUsage\n-----\n\nPyHive\n~~~~~~\n\nThe ``connect()`` function returns a ``pyhive`` Hive connection object, which internally wraps a ``thrift`` connection.\n\nConnecting with ``http_path``, ``host``, and a ``token``:\n\n.. code-block:: python\n\n    import os\n\n    from databricks_dbapi import hive\n\n\n    token = os.environ["DATABRICKS_TOKEN"]\n    host = os.environ["DATABRICKS_HOST"]\n    http_path = os.environ["DATABRICKS_HTTP_PATH"]\n\n\n    connection = hive.connect(\n        host=host,\n        http_path=http_path,\n        token=token,\n    )\n    cursor = connection.cursor()\n\n    cursor.execute("SELECT * FROM some_table LIMIT 100")\n\n    print(cursor.fetchone())\n    print(cursor.fetchall())\n\n\nThe ``pyhive`` connection also provides async functionality:\n\n.. code-block:: python\n\n    import os\n\n    from databricks_dbapi import hive\n    from TCLIService.ttypes import TOperationState\n\n\n    token = os.environ["DATABRICKS_TOKEN"]\n    host = os.environ["DATABRICKS_HOST"]\n    cluster = os.environ["DATABRICKS_CLUSTER"]\n\n\n    connection = hive.connect(\n        host=host,\n        cluster=cluster,\n        token=token,\n    )\n    cursor = connection.cursor()\n\n    cursor.execute("SELECT * FROM some_table LIMIT 100", async_=True)\n\n    status = cursor.poll().operationState\n    while status in (TOperationState.INITIALIZED_STATE, TOperationState.RUNNING_STATE):\n        logs = cursor.fetch_logs()\n        for message in logs:\n            print(message)\n\n        # If needed, an asynchronous query can be cancelled at any time with:\n        # cursor.cancel()\n\n        status = cursor.poll().operationState\n\n    print(cursor.fetchall())\n\n\nODBC\n~~~~\n\nThe ODBC DBAPI requires the Simba ODBC driver.\n\nConnecting with ``http_path``, ``host``, and a ``token``:\n\n.. code-block:: python\n\n    import os\n\n    from databricks_dbapi import odbc\n\n\n    token = os.environ["DATABRICKS_TOKEN"]\n    host = os.environ["DATABRICKS_HOST"]\n    http_path = os.environ["DATABRICKS_HTTP_PATH"]\n\n\n    connection = odbc.connect(\n        host=host,\n        http_path=http_path,\n        token=token,\n        driver_path="/path/to/simba/driver",\n    )\n    cursor = connection.cursor()\n\n    cursor.execute("SELECT * FROM some_table LIMIT 100")\n\n    print(cursor.fetchone())\n    print(cursor.fetchall())\n\n\nSQLAlchemy Dialects\n-------------------\n\ndatabricks+pyhive\n~~~~~~~~~~~~~~~~~\n\nInstalling registers the ``databricks+pyhive`` dialect/driver with SQLAlchemy. Fill in the required information when passing the engine URL.\n\n.. code-block:: python\n\n    from sqlalchemy import *\n    from sqlalchemy.engine import create_engine\n    from sqlalchemy.schema import *\n\n\n    engine = create_engine(\n        "databricks+pyhive://token:<databricks_token>@<host>:<port>/<database>",\n        connect_args={"http_path": "<cluster_http_path>"}\n    )\n\n    logs = Table("my_table", MetaData(bind=engine), autoload=True)\n    print(select([func.count("*")], from_obj=logs).scalar())\n\n\ndatabricks+pyodbc\n~~~~~~~~~~~~~~~~~\n\nInstalling registers the ``databricks+pyodbc`` dialect/driver with SQLAlchemy. Fill in the required information when passing the engine URL.\n\n.. code-block:: python\n\n    from sqlalchemy import *\n    from sqlalchemy.engine import create_engine\n    from sqlalchemy.schema import *\n\n\n    engine = create_engine(\n        "databricks+pyodbc://token:<databricks_token>@<host>:<port>/<database>",\n        connect_args={"http_path": "<cluster_http_path>", "driver_path": "/path/to/simba/driver"}\n    )\n\n    logs = Table("my_table", MetaData(bind=engine), autoload=True)\n    print(select([func.count("*")], from_obj=logs).scalar())\n\n\nRefer to the following documentation for more details on hostname, cluster name, and http path:\n\n* `Databricks <https://docs.databricks.com/user-guide/bi/jdbc-odbc-bi.html>`__\n* `Azure Databricks <https://docs.azuredatabricks.net/user-guide/bi/jdbc-odbc-bi.html>`__\n\n\nRelated\n-------\n\n* `pyhive <https://github.com/dropbox/PyHive>`__\n* `thrift <https://github.com/apache/thrift/tree/master/lib/py>`__\n* `pyodbc <https://github.com/mkleehammer/pyodbc>`__\n',
    'author': 'Christopher Flynn',
    'author_email': 'crf204@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/crflynn/databricks-dbapi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
}


setup(**setup_kwargs)



from databricks_dbapi import hive
from databricks_dbapi.sqlalchemy_dialects.base import DatabricksDialectBase


class DatabricksPyhiveDialect(DatabricksDialectBase):
    name = b"databricks"
    driver = b"pyhive"

    @classmethod
    def dbapi(cls):
        return hive

    def create_connect_args(self, url):
        kwargs = {
            "host": url.host,
            "port": url.port or 443,
            "user": url.username,
            "password": url.password,
            "database": url.database or "default",
        }

        if url.query is not None and "http_path" in url.query:
            kwargs["http_path"] = url.query["http_path"]

        kwargs.update(url.query)
        return [], kwargs



from ._version import __author__
from ._version import __author_email__
from ._version import __copyright__
from ._version import __description__
from ._version import __license__
from ._version import __title__
from ._version import __url__
from ._version import __version__



"""Version information."""
__title__ = "databricks_dbapi"
__description__ = "Databricks DBAPI."
__url__ = "https://github.com/crflynn/databricks-dbapi"
__version__ = "0.5.0"
__author__ = "Christopher Flynn"
__author_email__ = "crf204@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2018-2021 Christopher Flynn"



"""Provide a function to create a Hive connection to a Databricks cluster."""
import base64
import sys

from pyhive import hive
from pyhive.exc import *  # Make all exceptions visible in this module per PEP 249
from thrift.transport import THttpClient

PY_MAJOR = sys.version_info[0]

# PEP 249 module globals
apilevel = hive.apilevel
threadsafety = hive.threadsafety
paramstyle = hive.paramstyle


def connect(host, port=443, database="default", http_path=None, token=None, user=None, password=None, **kwargs):
    """Create a pyhive-driven DBAPI connection to Databricks workspace cluster.

    Create a DBAPI connection to a Databricks cluster, which can be used to generate
    DBAPI cursor(s). Provide an ``http_path`` from the cluster's JDBC/ODBC connection
    details.

    For authentication, provide either a ``token`` OR both a ``user`` and ``password``.
    Token authentication is strongly recommended over passwords.

    The simplest connection requires providing args ``host``, ``http_path``, and
    ``token``.

    :param str host: the server hostname from the cluster's JDBC/ODBC connection page.
    :param int port: the port number from the cluster's JDBC/ODBC connection page.
    :param str database: the database to use
    :param str http_path: the HTTP Path as shown in the cluster's JDBC/ODBC connection
        page.
    :param str token: a Databricks API token.
    :param str user: a Databricks user name.
    :param str password: the corresponding Databricks user's password.
    :param dict kwargs: keyword arguments passed to ``hive.connect``
    """
    if token is not None:
        auth = "token:%s" % token
    elif user is not None and password is not None:
        auth = "%s:%s" % (user, password)
    else:
        raise ValueError("Missing arguments. Must provide either token or user/password.")

    # https://kb.databricks.com/python/python-2-eol.html
    if PY_MAJOR < 3:
        auth = base64.standard_b64encode(auth)
    else:
        auth = base64.standard_b64encode(auth.encode()).decode()

    if http_path is not None:
        url = "https://%s:%s/%s" % (host, port, http_path)
    else:
        raise ValueError("Missing arguments. Must provide either cluster or http_path.")

    transport = THttpClient.THttpClient(url)
    transport.setCustomHeaders({"Authorization": "Basic %s" % auth})

    return hive.connect(database=database, thrift_transport=transport, **kwargs)



import re
from abc import ABC

from pyhive.sqlalchemy_hive import HiveDialect
from pyhive.sqlalchemy_hive import _type_map
from sqlalchemy import types
from sqlalchemy import util


class DatabricksDialectBase(HiveDialect, ABC):
    def get_table_names(self, connection, schema=None, **kw):
        query = "SHOW TABLES"
        if schema:
            query += " IN " + self.identifier_preparer.quote_identifier(schema)
        return [row[1] for row in connection.execute(query)]

    def get_columns(self, connection, table_name, schema=None, **kw):
        """Get columns according to Databricks' hive or oss hive."""
        rows = self._get_table_columns(connection, table_name, schema)
        # Strip whitespace
        rows = [[col.strip() if col else None for col in row] for row in rows]
        # Filter out empty rows and comment
        rows = [row for row in rows if row[0] and row[0] != "# col_name"]
        result = []
        for (col_name, col_type, _comment) in rows:
            # Handle both oss hive and Databricks' hive partition header, respectively
            if col_name in ("# Partition Information", "# Partitioning"):
                break
            # Take out the more detailed type information
            # e.g. 'map<int,int>' -> 'map'
            #      'decimal(10,1)' -> decimal
            col_type = re.search(r"^\w+", col_type).group(0)
            try:
                coltype = _type_map[col_type]
            except KeyError:
                util.warn("Did not recognize type '%s' of column '%s'" % (col_type, col_name))
                coltype = types.NullType

            result.append(
                {
                    "name": col_name,
                    "type": coltype,
                    "nullable": True,
                    "default": None,
                }
            )
        return result
